#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Implements adapters for time series annotation."""

__all__ = ["PyODAnnotator"]

from sktime.annotation.adapters._pyod import PyODAnnotator
