# -*- coding: utf-8 -*-
"""Early classification time series classifiers."""
__all__ = [
    "ProbabilityThresholdEarlyClassifier",
]

from sktime.classification.early_classification._probability_threshold import (
    ProbabilityThresholdEarlyClassifier,
)
