# -*- coding: utf-8 -*-
"""Averaging metrics for clustering."""
__all__ = ["mean_average"]
__author__ = ["chrisholder", "TonyBagnall"]

from sktime.clustering.metrics.averaging._averaging import mean_average
